import json

import pytest
import responses

from datarobot import DatetimeModel, enums, errors, Model
from datarobot.models.feature_effect import (
    FeatureEffectMetadata,
    FeatureEffectMetadataDatetime,
    FeatureEffects,
)


@pytest.fixture
def feature_effect_metadata():
    return {"status": "COMPLETED", "sources": ["training", "validation", "holdout"]}


@pytest.fixture
def feature_effect_metadata_datetime():
    return {
        "data": [
            {"backtestIndex": "0", "status": "COMPLETED", "sources": ["training", "validation"]},
            {
                "backtestIndex": "holdout",
                "status": "NOT_COMPLETED",
                "sources": ["training", "holdout"],
            },
        ]
    }


@pytest.fixture
def feature_effect_server_data():
    return {
        "projectId": "project_id",
        "modelId": "model_id",
        "source": "training",
        "featureEffects": [
            {
                "featureType": "numeric",
                "predictedVsActual": {
                    "isCapped": False,
                    "data": [
                        {
                            "rowCount": 46.5,
                            "actual": 16,
                            "predicted": 15,
                            "label": "[ 1872, 1879 )",
                            "bin": ["1872", "1879"],
                        },
                        {
                            "rowCount": 31.5,
                            "actual": 752,
                            "predicted": 799.43,
                            "label": "[ 1879, 1886 )",
                            "bin": ["1879", "1886"],
                        },
                        {
                            "rowCount": 0.0,
                            "actual": None,
                            "predicted": None,
                            "label": "[ 1879, 1886 )",
                            "bin": ["1879", "1886"],
                        },
                    ],
                },
                "partialDependence": {
                    "isCapped": False,
                    "data": [
                        {"dependence": 41.25, "label": "1999"},
                        {"dependence": 40.64, "label": "1928"},
                        {"dependence": 41.44, "label": "nan"},
                    ],
                },
                "featureName": "record_min_temp_year",
                "weightLabel": None,
                "featureImpactScore": 1,
            },
            {
                "featureType": "categorical",
                "predictedVsActual": {
                    "isCapped": False,
                    "data": [
                        {"rowCount": 99, "actual": 4107, "predicted": 4110.0, "label": "1"},
                        {"rowCount": 98, "actual": 4175, "predicted": 4119.0, "label": "0"},
                    ],
                },
                "partialDependence": {
                    "isCapped": False,
                    "data": [
                        {"dependence": 41.13, "label": "1"},
                        {"dependence": 41.91, "label": "0"},
                        {"dependence": 41.92, "label": "=Other Unseen="},
                    ],
                },
                "featureName": "date (Day of Week)",
                "weightLabel": None,
                "featureImpactScore": 0.2,
            },
            {
                "featureType": "categorical",
                "partialDependence": {
                    "isCapped": False,
                    "data": [
                        {"dependence": 41.13, "label": "1"},
                        {"dependence": 41.91, "label": "0"},
                        {"dependence": 41.92, "label": "=Other Unseen="},
                    ],
                },
                "featureName": "date (Day of Week) no pvsa",
                "weightLabel": None,
                "featureImpactScore": 0.1,
            },
        ],
    }


@pytest.fixture
def feature_effect_server_data_holdout(feature_effect_server_data):
    feature_effect_server_data_holdout = dict(feature_effect_server_data)
    feature_effect_server_data_holdout["source"] = "holdout"
    return feature_effect_server_data_holdout


@pytest.fixture
def feature_effect_server_data_datetime(feature_effect_server_data):

    feature_effect_server_data_datetime = dict(feature_effect_server_data)
    feature_effect_server_data_datetime["backtestIndex"] = "0"
    return feature_effect_server_data_datetime


@pytest.fixture
def feature_effect_server_data_datetime_holdout(feature_effect_server_data_datetime):
    feature_effect_server_data_datetime_holdout = dict(feature_effect_server_data_datetime)
    feature_effect_server_data_datetime_holdout["source"] = "holdout"
    return feature_effect_server_data_datetime_holdout


@pytest.fixture
def feature_effect_response(feature_effect_server_data, feature_effect_url):
    source = "training"
    body = json.dumps(feature_effect_server_data)
    responses.add(
        responses.GET,
        "{}?source={}".format(feature_effect_url, source),
        status=200,
        content_type="application/json",
        body=body,
    )


@pytest.fixture
def feature_effect_response_holdout(feature_effect_server_data_holdout, feature_effect_url):
    source = "holdout"
    body = json.dumps(feature_effect_server_data_holdout)
    responses.add(
        responses.GET,
        "{}?source={}".format(feature_effect_url, source),
        status=200,
        content_type="application/json",
        body=body,
    )


@pytest.fixture
def feature_effect_invalid_request_response(feature_effect_url):
    source = "invalid"
    body = {"message": "Invalid source"}

    responses.add(
        responses.GET,
        "{}?source={}".format(feature_effect_url, source),
        status=404,
        content_type="application/json",
        body=json.dumps(body),
    )


@pytest.fixture
def feature_effect_response_datetime(
    feature_effect_server_data_datetime, feature_effect_url_datetime
):
    source = "training"
    backtest_index = "0"
    responses.add(
        responses.GET,
        "{}?source={}&backtestIndex=".format(feature_effect_url_datetime, source, backtest_index),
        status=200,
        content_type="application/json",
        body=json.dumps(feature_effect_server_data_datetime),
    )


@pytest.fixture
def feature_effect_response_holdout_datetime(
    feature_effect_server_data_datetime_holdout, feature_effect_url_datetime
):
    source = "training"
    backtest_index = "0"
    responses.add(
        responses.GET,
        "{}?source={}&backtestIndex=".format(feature_effect_url_datetime, source, backtest_index),
        status=200,
        content_type="application/json",
        body=json.dumps(feature_effect_server_data_datetime_holdout),
    )


@pytest.fixture
def feature_effect_invalid_request_response_datetime(feature_effect_url_datetime):
    body = {"message": "Invalid source"}

    source = "invalid"
    backtest_index = "0"
    responses.add(
        responses.GET,
        "{}?source={}&backtestIndex=".format(feature_effect_url_datetime, source, backtest_index),
        status=404,
        content_type="application/json",
        body=json.dumps(body),
    )


@pytest.fixture
def feature_effect_metadata_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/featureEffectsMetadata/".format(
        project_id, model_id
    )


@pytest.fixture
def feature_effect_metadata_url_datetime(project_id, model_id):
    return "https://host_name.com/projects/{}/datetimeModels/{}/featureEffectsMetadata/".format(
        project_id, model_id
    )


@pytest.fixture
def feature_effect_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/featureEffects/".format(
        project_id, model_id
    )


@pytest.fixture
def feature_effect_url_datetime(project_id, model_id):
    return "https://host_name.com/projects/{}/datetimeModels/{}/featureEffects/".format(
        project_id, model_id
    )


@pytest.fixture
def feature_effect_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.FEATURE_EFFECTS)


@pytest.fixture
def feature_effect_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.FEATURE_EFFECTS)


@pytest.fixture
def feature_effect_job_creation_response(feature_effect_url, job_url):
    responses.add(
        responses.POST,
        feature_effect_url,
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def feature_effect_job_creation_bad_request(feature_effect_url):
    body = {"message": "Invalid field data", "errors": {"rowCount": "value is less than 10"}}
    responses.add(
        responses.POST,
        feature_effect_url,
        status=422,
        content_type="application/json",
        body=json.dumps(body),
    )


@pytest.fixture
def feature_effect_job_creation_response_datetime(feature_effect_url_datetime, job_url):
    responses.add(
        responses.POST,
        feature_effect_url_datetime,
        body=json.dumps({"backtestIndex": "0"}),
        status=202,
        adding_headers={"Location": job_url},
        content_type="application/json",
    )


@pytest.fixture
def feature_effect_completed_response(
    feature_effect_job_finished_server_data, job_url, feature_effect_url
):
    """
    Loads a response that the given job is a featureImpact job, and is in completed
    """
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_effect_job_finished_server_data),
        status=303,
        adding_headers={"Location": feature_effect_url},
        content_type="application/json",
    )


@pytest.fixture
def feature_effect_completed_response_datetime(
    feature_effect_job_finished_server_data, job_url, feature_effect_url_datetime
):
    """
    Loads a response that the given job is a featureImpact job, and is in completed
    """
    feature_effect_url_datetime_with_backtest = "{}?backtestIndex={}".format(
        feature_effect_url_datetime, "0"
    )
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_effect_job_finished_server_data),
        status=303,
        adding_headers={"Location": feature_effect_url_datetime_with_backtest},
        content_type="application/json",
    )


@pytest.fixture
def feature_effect_running_response(feature_effect_job_running_server_data, job_url):
    """
    Loads a response that the given job is a featureEffects job, and is running
    """
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_effect_job_running_server_data),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def feature_effect_previously_ran_response(feature_effect_url, job_id):
    """
    Loads a response that the given model has already ran its feature effect
     """
    body = {
        "message": "Feature Effect is in progress for this model.",
        "errorName": "JobAlreadyAdded",
        "jobId": job_id,
    }
    responses.add(
        responses.POST,
        feature_effect_url,
        body=json.dumps(body),
        status=422,
        content_type="application/json",
    )


@pytest.fixture
def feature_effect_previously_ran_response_datetime(feature_effect_url_datetime, job_id):
    """
    Loads a response that the given model has already ran its feature effect
     """
    body = {
        "message": "Feature Effect is in progress for this model.",
        "errorName": "JobAlreadyAdded",
        "jobId": job_id,
    }
    responses.add(
        responses.POST,
        feature_effect_url_datetime,
        body=json.dumps(body),
        status=422,
        content_type="application/json",
    )


def test_get_feature_effect_metadata_url(project_id, model_id):
    model = Model(id=model_id, project_id=project_id)
    expected_fe_meatadata_url = "projects/{}/models/{}/featureEffectsMetadata/".format(
        project_id, model_id
    )
    assert model._get_feature_effect_metadata_url() == expected_fe_meatadata_url


def test_get_feature_effect_metadata_url_datetime(project_id, model_id):
    model = DatetimeModel(id=model_id, project_id=project_id)
    expected_fe_meatadata_url = "projects/{}/datetimeModels/{}/featureEffectsMetadata/".format(
        project_id, model_id
    )
    assert model._get_feature_effect_metadata_url() == expected_fe_meatadata_url


@responses.activate
def test_get_feature_effect_metadata(
    feature_effect_metadata, feature_effect_metadata_url, project_id, model_id
):
    responses.add(
        responses.GET,
        feature_effect_metadata_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_effect_metadata),
    )
    model = Model(id=model_id, project_id=project_id)
    fe_metadata = model.get_feature_effect_metadata()

    assert isinstance(fe_metadata, FeatureEffectMetadata)
    assert fe_metadata.status == feature_effect_metadata["status"]
    assert fe_metadata.sources == feature_effect_metadata["sources"]


@responses.activate
def test_get_feature_effect_metadata_datetime(
    feature_effect_metadata_datetime, feature_effect_metadata_url_datetime, project_id, model_id,
):
    responses.add(
        responses.GET,
        feature_effect_metadata_url_datetime,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_effect_metadata_datetime),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    fe_metadata = model.get_feature_effect_metadata()

    assert isinstance(fe_metadata, FeatureEffectMetadataDatetime)
    assert len(fe_metadata.data) == len(feature_effect_metadata_datetime["data"])
    feature_effect_metadata_datetime_sorted = sorted(
        feature_effect_metadata_datetime["data"], key=lambda k: k["backtestIndex"]
    )
    for i, meta in enumerate(sorted(fe_metadata)):
        assert meta.backtest_index == feature_effect_metadata_datetime_sorted[i]["backtestIndex"]
        assert meta.status == feature_effect_metadata_datetime_sorted[i]["status"]
        assert sorted(meta.sources) == sorted(feature_effect_metadata_datetime_sorted[i]["sources"])


@responses.activate
@pytest.mark.usefixtures("feature_effect_job_creation_response", "feature_effect_running_response")
def test_get_feature_effect_job_result_not_finished(one_model):
    feature_effect_job = one_model.request_feature_effect()
    with pytest.raises(errors.JobNotFinished):
        feature_effect_job.get_result()


@responses.activate
@pytest.mark.usefixtures("feature_effect_job_creation_bad_request")
def test_request_feature_effect_bad_payload(one_model):
    with pytest.raises(errors.ClientError):
        one_model.request_feature_effect(row_count=0)


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_job_creation_response",
    "feature_effect_completed_response",
    "feature_effect_response",
)
def test_get_feature_effect_job_result_finished(feature_effect_server_data, one_model):
    feature_effect_job = one_model.request_feature_effect(row_count=3200)
    feature_effect = feature_effect_job.get_result()
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert feature_effect.model_id == feature_effect_server_data["modelId"]
    assert feature_effect.source == feature_effect_server_data["source"]
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert len(feature_effect.feature_effects) == len(feature_effect_server_data["featureEffects"])

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_response_holdout",
    "feature_effect_job_creation_response",
    "feature_effect_completed_response",
)
def test_get_feature_effect_job_result_finished_holdout(
    feature_effect_server_data_holdout, one_model
):
    feature_effect_job = one_model.request_feature_effect()
    params = {"source": "holdout"}
    feature_effect = feature_effect_job.get_result(params)
    assert feature_effect.project_id == feature_effect_server_data_holdout["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_holdout["modelId"]
    assert feature_effect.source == feature_effect_server_data_holdout["source"]
    assert feature_effect.project_id == feature_effect_server_data_holdout["projectId"]
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_holdout["featureEffects"]
    )

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_holdout)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures("feature_effect_job_creation_response", "feature_effect_running_response")
def test_wait_for_feature_effect_never_finished(one_model, mock_async_time):
    mock_async_time.time.side_effect = (0, 5)
    feature_effect_job = one_model.request_feature_effect()
    with pytest.raises(errors.AsyncTimeoutError):
        feature_effect_job.get_result_when_complete(max_wait=1)


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_previously_ran_response",
    "feature_effect_completed_response",
    "feature_effect_response",
)
def test_get_or_request_feature_effect_previously_requested(one_model, feature_effect_server_data):
    feature_effect = one_model.get_or_request_feature_effect(source="training")
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert feature_effect.model_id == feature_effect_server_data["modelId"]
    assert feature_effect.source == feature_effect_server_data["source"]
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert len(feature_effect.feature_effects) == len(feature_effect_server_data["featureEffects"])

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_previously_ran_response", "feature_effect_running_response"
)
def test_get_or_request_feature_effect_currently_running_waits(one_model, mock_async_time):
    mock_async_time.time.side_effect = (0, 5)
    with pytest.raises(errors.AsyncTimeoutError):
        one_model.get_or_request_feature_effect(max_wait=1, source="training")


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_job_creation_response",
    "feature_effect_completed_response",
    "feature_effect_response",
)
def test_get_or_request_feature_effect(one_model, feature_effect_server_data):

    feature_effect = one_model.get_or_request_feature_effect(source="training", row_count=3200)
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert feature_effect.model_id == feature_effect_server_data["modelId"]
    assert feature_effect.source == feature_effect_server_data["source"]
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert len(feature_effect.feature_effects) == len(feature_effect_server_data["featureEffects"])

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures("feature_effect_job_creation_bad_request")
def test_get_or_request_feature_effect_bad_payload(one_model):
    with pytest.raises(errors.ClientError):
        one_model.get_or_request_feature_effect(source="training", row_count=1000000)


@responses.activate
@pytest.mark.usefixtures(
    "client",
    "feature_effect_job_creation_response",
    "feature_effect_completed_response",
    "feature_effect_response",
)
def test_wait_for_feature_effect_finished(one_model, feature_effect_server_data):
    params = {"source": "training"}
    feature_effect_job = one_model.request_feature_effect()
    feature_effect = feature_effect_job.get_result_when_complete(max_wait=0.5, params=params)

    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert feature_effect.model_id == feature_effect_server_data["modelId"]
    assert feature_effect.source == feature_effect_server_data["source"]
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert len(feature_effect.feature_effects) == len(feature_effect_server_data["featureEffects"])

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures("feature_effect_response")
def test_get_feature_effect_assumed_complete(one_model, feature_effect_server_data):
    feature_effect = one_model.get_feature_effect(source="training")

    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert feature_effect.model_id == feature_effect_server_data["modelId"]
    assert feature_effect.source == feature_effect_server_data["source"]
    assert feature_effect.project_id == feature_effect_server_data["projectId"]
    assert len(feature_effect.feature_effects) == len(feature_effect_server_data["featureEffects"])

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures("feature_effect_invalid_request_response")
def test_get_feature_effect_invalid_source(one_model):
    with pytest.raises(errors.ClientError):
        one_model.get_feature_effect(source="invalid")


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_job_creation_response_datetime", "feature_effect_running_response"
)
def test_get_feature_effect_job_result_not_finished_datetime(one_datetime_model):
    backtest_index = "0"
    feature_effect_job = one_datetime_model.request_feature_effect(backtest_index)
    with pytest.raises(errors.JobNotFinished):
        feature_effect_job.get_result()


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_job_creation_response_datetime",
    "feature_effect_completed_response_datetime",
    "feature_effect_response_datetime",
)
def test_get_feature_effect_job_result_finished_datetime(
    feature_effect_server_data_datetime, one_datetime_model
):
    backtest_index = "0"
    source = "training"
    params = {"source": source}
    feature_effect_job = one_datetime_model.request_feature_effect(backtest_index)
    feature_effect = feature_effect_job.get_result(params)
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_datetime["modelId"]
    assert feature_effect.source == feature_effect_server_data_datetime["source"]
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.backtest_index == feature_effect_server_data_datetime["backtestIndex"]
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_datetime["featureEffects"]
    )

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_datetime)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_response_holdout_datetime",
    "feature_effect_job_creation_response_datetime",
    "feature_effect_completed_response_datetime",
)
def test_get_feature_effect_job_result_finished_holdout_datetime(
    feature_effect_server_data_datetime_holdout, one_datetime_model
):
    backtest_index = "0"
    feature_effect_job = one_datetime_model.request_feature_effect(backtest_index)
    params = {"source": "holdout"}
    feature_effect = feature_effect_job.get_result(params)
    assert feature_effect.project_id == feature_effect_server_data_datetime_holdout["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_datetime_holdout["modelId"]
    assert feature_effect.source == feature_effect_server_data_datetime_holdout["source"]
    assert feature_effect.project_id == feature_effect_server_data_datetime_holdout["projectId"]
    assert (
        feature_effect.backtest_index
        == feature_effect_server_data_datetime_holdout["backtestIndex"]
    )
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_datetime_holdout["featureEffects"]
    )

    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_datetime_holdout)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_job_creation_response_datetime", "feature_effect_running_response"
)
def test_wait_for_feature_effect_never_finished_datetime(one_datetime_model, mock_async_time):
    backtest_index = "0"
    mock_async_time.time.side_effect = (0, 5)
    feature_effect_job = one_datetime_model.request_feature_effect(backtest_index)
    with pytest.raises(errors.AsyncTimeoutError):
        feature_effect_job.get_result_when_complete(max_wait=1)


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_previously_ran_response_datetime",
    "feature_effect_completed_response_datetime",
    "feature_effect_response_datetime",
)
def test_get_or_request_feature_effect_previously_requested_datetime(
    one_datetime_model, feature_effect_server_data_datetime
):
    backtest_index = "0"
    source = "training"
    feature_effect = one_datetime_model.get_or_request_feature_effect(
        source=source, backtest_index=backtest_index
    )
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_datetime["modelId"]
    assert feature_effect.source == feature_effect_server_data_datetime["source"]
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_datetime["featureEffects"]
    )
    assert feature_effect.backtest_index == feature_effect_server_data_datetime["backtestIndex"]
    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_datetime)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_previously_ran_response_datetime", "feature_effect_running_response"
)
def test_get_or_request_feature_effect_currently_running_waits_datetime(
    one_datetime_model, mock_async_time
):
    backtest_index = "0"
    source = "training"
    mock_async_time.time.side_effect = (0, 5)
    with pytest.raises(errors.AsyncTimeoutError):
        one_datetime_model.get_or_request_feature_effect(
            source=source, backtest_index=backtest_index, max_wait=1
        )


@responses.activate
@pytest.mark.usefixtures(
    "feature_effect_job_creation_response_datetime",
    "feature_effect_completed_response_datetime",
    "feature_effect_response_datetime",
)
def test_get_or_request_feature_effect_datetime(
    one_datetime_model, feature_effect_server_data_datetime
):
    backtest_index = "0"
    source = "training"
    feature_effect = one_datetime_model.get_or_request_feature_effect(
        source=source, backtest_index=backtest_index
    )
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_datetime["modelId"]
    assert feature_effect.source == feature_effect_server_data_datetime["source"]
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_datetime["featureEffects"]
    )
    assert feature_effect.backtest_index == feature_effect_server_data_datetime["backtestIndex"]
    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_datetime)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures(
    "client",
    "feature_effect_job_creation_response_datetime",
    "feature_effect_completed_response_datetime",
    "feature_effect_response_datetime",
)
def test_wait_for_feature_effect_finished_datetime(
    one_datetime_model, feature_effect_server_data_datetime
):
    params = {"source": "training"}
    backtest_index = "0"
    feature_effect_job = one_datetime_model.request_feature_effect(backtest_index)
    feature_effect = feature_effect_job.get_result_when_complete(max_wait=0.5, params=params)
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_datetime["modelId"]
    assert feature_effect.source == feature_effect_server_data_datetime["source"]
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.backtest_index == feature_effect_server_data_datetime["backtestIndex"]
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_datetime["featureEffects"]
    )
    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_datetime)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures("feature_effect_response_datetime")
def test_get_feature_effect_assumed_complete_datetime(
    one_datetime_model, feature_effect_server_data_datetime
):
    backtest_index = "0"
    source = "invalid"
    feature_effect = one_datetime_model.get_feature_effect(
        source=source, backtest_index=backtest_index
    )
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.model_id == feature_effect_server_data_datetime["modelId"]
    assert feature_effect.source == feature_effect_server_data_datetime["source"]
    assert feature_effect.project_id == feature_effect_server_data_datetime["projectId"]
    assert feature_effect.backtest_index == feature_effect_server_data_datetime["backtestIndex"]
    assert len(feature_effect.feature_effects) == len(
        feature_effect_server_data_datetime["featureEffects"]
    )
    expected_fe = FeatureEffects.from_server_data(feature_effect_server_data_datetime)
    assert expected_fe == feature_effect


@responses.activate
@pytest.mark.usefixtures("feature_effect_invalid_request_response_datetime")
def test_get_feature_effect_invalid_source_datetime(one_datetime_model):
    backtest_index = "0"
    source = "invalid"
    with pytest.raises(errors.ClientError):
        one_datetime_model.get_feature_effect(source=source, backtest_index=backtest_index)



import trafaret as t

from datarobot import CustomModelVersion, ExecutionEnvironment, ExecutionEnvironmentVersion
from datarobot._experimental.models.custom_training_model import CustomTrainingModel
from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate


class CustomTrainingBlueprint(APIObject):
    """A custom training blueprint.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        user blueprint id
    custom_model_version: dict
        dict with 2 keys: `id` and `label`,
        where `id` is the ID of the custom model version
        and `label` is the version label
    training_history: list
        List of dicts with 6 keys:
        where `pid` is the ID of the project
        and `project_name` is the name of the project
        and `lid` is the leaderboard id
        and `creation_date` is a ISO-8601 timestamp of when the project
          the blueprint was trained on was created
        and `project_models_count` is the number of models in the project
          that the blueprint was trained on
        and `target_name` is the name of the project's target
    """

    _path = "customTrainingBlueprints/"
    _converter = t.Dict(
        {
            t.Key("user_blueprint_id") >> "id": t.String(),
            t.Key("custom_model_version"): t.Dict(
                {t.Key("id"): t.String(), t.Key("label"): t.String()}
            ),
            t.Key("training_history"): t.List(
                t.Dict(
                    {
                        t.Key("pid"): t.String(),
                        t.Key("project_name"): t.String(),
                        t.Key("lid"): t.String(),
                        t.Key("creation_date"): t.String(),
                        t.Key("project_models_count"): t.Int(),
                        t.Key("target_name"): t.String(),
                    }
                )
            ),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return encode_utf8_if_py2(u"{}({!r})".format(self.__class__.__name__, self.id))

    def _set_values(
        self, id, custom_model_version, training_history, project_id=None, filenames=None,
    ):
        self.id = id
        self.custom_model_version = custom_model_version
        self.training_history = training_history
        self.project_id = project_id
        self.filenames = filenames

    @classmethod
    def create(
        cls, custom_model_version_id=None, custom_model_id=None,
    ):
        """Create a custom training blueprint.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_version_id: Optional[str]
            the id of the custom model version
        custom_model_id: Optional[str]
            the id of the custom model


        if custom_model_id is provided, latest version will be used.

        Returns
        -------
        CustomTrainingBlueprint
            created custom training blueprint

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if not custom_model_version_id:
            if not custom_model_id:
                raise ValueError(u"Must provide custom_model_version_id or custom_model_id")

            model = CustomTrainingModel.get(custom_model_id=custom_model_id)
            custom_model_version_id = model.latest_version.id
        payload = {
            "custom_model_version_id": custom_model_version_id,
        }
        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def create_from_scratch(cls, name, environment_dir, training_code_files, target_type):
        desc = "Generated from python client"
        # Make environment
        ee = ExecutionEnvironment.create(
            name=name + "-environment", description=desc, programming_language="python"
        )
        ExecutionEnvironmentVersion.create(
            str(ee.id), environment_dir, description=desc,
        )

        # Make custom model
        cm = CustomTrainingModel.create(
            name=name + "-model", target_type=target_type, description=desc
        )
        cmv = CustomModelVersion.create_clean(
            custom_model_id=cm.id, base_environment_id=ee.id, files=training_code_files
        )

        blueprint = cls.create(custom_model_version_id=cmv.id)
        blueprint.filenames = training_code_files
        return blueprint

    @classmethod
    def create_from_dropin(
        cls, model_name, dropin_env_id, target_type, training_code_files=None, folder_path=None
    ):
        new_custom_training_model = CustomTrainingModel.create(
            name=model_name, target_type=target_type
        )
        cmv = CustomModelVersion.create_clean(
            new_custom_training_model.id,
            base_environment_id=dropin_env_id,
            files=training_code_files,
            folder_path=folder_path,
        )
        blueprint = cls.create(custom_model_version_id=cmv.id)
        blueprint.filenames = training_code_files
        return blueprint

    @classmethod
    def list(cls):
        """List custom training blueprints.

        .. versionadded:: v2.21

        Returns
        -------
        List[CustomTrainingBlueprint]
            a list of custom learning blueprints

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(cls._path, {}, cls._client)
        return [cls.from_server_data(item) for item in data]



import pandas as pd
import six
import trafaret as t

from datarobot import errors
from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2, get_id_from_response
from datarobot.utils.pagination import unpaginate


class ShapMatrix(APIObject):
    """
    Represents SHAP based prediction explanations and provides access to score values.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    shap_matrix_id : str
        id of the generated SHAP matrix
    model_id : str
        id of the model used to
    dataset_id : str
         id of the prediction dataset SHAP values were computed for

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # request SHAP matrix calculation
        shap_matrix_job = dr.ShapMatrix.create(project_id, model_id, dataset_id)
        shap_matrix = shap_matrix_job.get_result_when_complete()

        # list available SHAP matrices
        shap_matrices = dr.ShapMatrix.list(project_id)
        shap_matrix = shap_matrices[0]

        # get SHAP matrix as dataframe
        shap_matrix_values = shap_matrix.get_as_dataframe()
    """

    _path = "projects/{}/shapMatrices/"
    _converter = t.Dict(
        {
            t.Key("id"): t.String(),
            t.Key("project_id"): t.String(),
            t.Key("model_id"): t.String(),
            t.Key("dataset_id"): t.String(),
        }
    ).allow_extra("*")

    def __init__(self, project_id, id, model_id=None, dataset_id=None):
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.id = id

    def __repr__(self):
        template = u"{}(id={!r}, project_id={!r}, model_id={!r}, dataset_id={!r})"
        return encode_utf8_if_py2(
            template.format(
                type(self).__name__, self.id, self.project_id, self.model_id, self.dataset_id,
            )
        )

    @classmethod
    def create(cls, project_id, model_id, dataset_id):
        """ Calculate SHAP based prediction explanations against previously uploaded dataset.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which prediction explanations are requested
        dataset_id : str
            id of the prediction dataset for which prediction explanations are requested (as
            uploaded from Project.upload_dataset)

        Returns
        -------
        job : ShapMatrixJob
            The job computing the SHAP based prediction explanations

        Raises
        ------
        ClientError
            If the server responded with 4xx status. Possible reasons are project, model or dataset
            don't exist, user is not allowed or model doesn't support SHAP based prediction
            explanations
        ServerError
            If the server responded with 5xx status
        """
        data = {"model_id": model_id, "dataset_id": dataset_id}
        url = "projects/{}/shapMatrices/".format(project_id)
        response = cls._client.post(url, data=data)
        job_id = get_id_from_response(response)
        from .shap_matrix_job import ShapMatrixJob

        return ShapMatrixJob.get(
            project_id=project_id, job_id=job_id, model_id=model_id, dataset_id=dataset_id
        )

    @classmethod
    def from_location(cls, location, model_id=None, dataset_id=None):
        head, tail = location.split("/shapMatrices/", 1)
        project_id, id = head.split("/")[-1], tail.split("/")[0]
        return cls(project_id=project_id, id=id, model_id=model_id, dataset_id=dataset_id)

    @classmethod
    def list(cls, project_id):
        """
        Fetch all the computed SHAP prediction explanations for a project.

        Parameters
        ----------
        project_id : str
            id of the project

        Returns
        -------
        List of ShapMatrix
            A list of :py:class:`ShapMatrix <datarobot.models.ShapMatrix>` objects

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            initial_url=cls._path.format(project_id), initial_params=None, client=cls._client
        )
        result = [cls.from_server_data(item) for item in data]
        return result

    @classmethod
    def get(cls, project_id, id):
        """
        Retrieve the specific SHAP matrix.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        id : str
            id of the SHAP matrix

        Returns
        -------
        :py:class:`ShapMatrix <datarobot.models.ShapMatrix>` object representing specified record
        """
        return cls(project_id=project_id, id=id)

    def get_as_dataframe(self):
        """
        Retrieve SHAP matrix values as dataframe.

        Returns
        -------
        dataframe : pandas.DataFrame
            A dataframe with SHAP scores

        Raises
        ------
        datarobot.dse.errors.ClientError
            if the server responded with 4xx status.
        datarobot.dse.errors.ServerError
            if the server responded with 5xx status.
        """
        path = self._path.format(self.project_id) + "{}/".format(self.id)
        resp = self._client.get(path, headers={"Accept": "text/csv"}, stream=True)
        if resp.status_code == 200:
            content = resp.content.decode("utf-8")
            return pd.read_csv(six.StringIO(content), index_col=0, encoding="utf-8")
        else:
            raise errors.ServerError(
                "Server returned unknown status code: {}".format(resp.status_code),
                resp.status_code,
            )



from collections import OrderedDict
import json

import dateutil
import pytest
import responses

from datarobot.models import Deployment, ServiceStats, ServiceStatsOverTime
from datarobot.utils import from_api


class TestServiceStats(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "period": {
                "start": "2019-08-01T00:00:00.000000Z",
                "end": "2019-08-10T00:00:00.000000Z",
            },
            "metrics": {"totalPredictions": 1000, "executionTime": 50.27, "responseTime": None},
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = "{}/deployments/{}/serviceStats/".format(unittest_endpoint, deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_service_stats(self, deployment_data, response_data):
        service_stats_objs = [
            ServiceStats.get(deployment_data["id"]),
            Deployment.get(deployment_data["id"]).get_service_stats(),
        ]
        for service_stats in service_stats_objs:
            assert service_stats.model_id == response_data["model_id"]

            expected_start_date = dateutil.parser.parse(response_data["period"]["start"])
            expected_end_date = dateutil.parser.parse(response_data["period"]["end"])
            assert service_stats.period["start"] == expected_start_date
            assert service_stats.period["end"] == expected_end_date

            assert service_stats.metrics.keys() == response_data["metrics"].keys()
            for metric_name, metric in service_stats.metrics.items():
                expected_value = response_data["metrics"][metric_name]
                assert metric == expected_value
                assert service_stats[metric_name] == expected_value

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    @pytest.mark.parametrize(
        "kwarg, expected_query_string",
        [
            ({"model_id": "abc"}, "?modelId=abc"),
            ({"execution_time_quantile": 0.74}, "?executionTimeQuantile=0.74"),
            ({"response_time_quantile": 0.43}, "?responseTimeQuantile=0.43"),
            ({"slow_requests_threshold": 40}, "?slowRequestsThreshold=40"),
        ],
    )
    def test_params(self, deployment_data, kwarg, expected_query_string):
        """Test optional params are correctly forwarded to the API"""

        ServiceStats.get(deployment_data["id"], **kwarg)
        assert expected_query_string in responses.calls[0].request.url

        Deployment.get(deployment_data["id"]).get_service_stats(**kwarg)
        assert expected_query_string in responses.calls[0].request.url


class TestServiceStatsOverTime(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "metric": "totalPredictions",
            "summary": {
                "period": {
                    "start": "2019-08-01T00:00:00.000000Z",
                    "end": "2019-08-04T00:00:00.000000Z",
                },
                "value": 2250,
            },
            "buckets": [
                {
                    "period": {
                        "start": "2019-08-01T00:00:00.000000Z",
                        "end": "2019-08-02T00:00:00.000000Z",
                    },
                    "value": 1250,
                },
                {
                    "period": {
                        "start": "2019-08-02T00:00:00.000000Z",
                        "end": "2019-08-03T00:00:00.000000Z",
                    },
                    "value": 1000,
                },
                {
                    "period": {
                        "start": "2019-08-03T00:00:00.000000Z",
                        "end": "2019-08-04T00:00:00.000000Z",
                    },
                    "value": 0,
                },
            ],
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        deployment_id = deployment_data["id"]
        url = "{}/deployments/{}/serviceStatsOverTime/".format(unittest_endpoint, deployment_id)
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    def assert_bucket(self, actual, expected):
        assert actual.keys() == expected.keys()

        actual_period = actual["period"]
        if actual_period:
            assert actual["period"].keys() == expected["period"].keys()
            assert actual_period["start"] == dateutil.parser.parse(expected["period"]["start"])
            assert actual_period["end"] == dateutil.parser.parse(expected["period"]["end"])
        else:
            assert actual_period is None

        assert actual["value"] == expected["value"]

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_service_stats_over_time(self, deployment_data, response_data):
        service_stats_over_time_objs = [
            ServiceStatsOverTime.get(deployment_data["id"]),
            Deployment.get(deployment_data["id"]).get_service_stats_over_time(),
        ]
        for service_stats_over_time in service_stats_over_time_objs:
            expected = from_api(response_data, keep_null_keys=True)
            assert service_stats_over_time.metric == expected["metric"]
            assert service_stats_over_time.model_id == expected["model_id"]

            self.assert_bucket(service_stats_over_time.summary, expected["summary"])
            for index, bucket in enumerate(service_stats_over_time.buckets):
                self.assert_bucket(bucket, expected["buckets"][index])

            expected_bucket_values = [
                (dateutil.parser.parse(bucket["period"]["start"]), bucket["value"])
                for bucket in expected["buckets"]
            ]
            assert service_stats_over_time.bucket_values == OrderedDict(expected_bucket_values)

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    @pytest.mark.parametrize(
        "kwarg, expected_query_string",
        [
            ({"metric": "LogLoss"}, "?metric=LogLoss"),
            ({"model_id": "abc"}, "?modelId=abc"),
            ({"bucket_size": "P1D"}, "?bucketSize=P1D"),
            ({"quantile": 0.74}, "?quantile=0.74"),
            ({"threshold": 0.43}, "?threshold=0.43"),
        ],
    )
    def test_params(self, deployment_data, kwarg, expected_query_string):
        """Test optional params are correctly forwarded to the API"""

        ServiceStatsOverTime.get(deployment_data["id"], **kwarg)
        assert expected_query_string in responses.calls[0].request.url

        Deployment.get(deployment_data["id"]).get_service_stats_over_time(**kwarg)
        assert expected_query_string in responses.calls[0].request.url



import json

import pytest
import responses
from six.moves.urllib.parse import urljoin

from datarobot import enums


@pytest.fixture
def model_url(project_url, model_id):
    return urljoin(project_url, "models/{}/".format(model_id))


@pytest.fixture
def feature_impact_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/featureImpact/".format(project_id, model_id)


@pytest.fixture
def multiclass_feature_impact_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/multiclassFeatureImpact/".format(
        project_id, model_id
    )


@pytest.fixture
def feature_impact_job_creation_response(feature_impact_url, job_url):
    responses.add(
        responses.POST,
        feature_impact_url,
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def multiclass_feature_impact_server_data():
    return {
        "ranRedundancyDetection": False,
        "classFeatureImpacts": [
            {
                "featureImpacts": [
                    {
                        "featureName": "Petal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.157,
                        "impactNormalized": 0.701,
                    },
                    {
                        "featureName": "Petal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.121,
                        "impactNormalized": 0.539,
                    },
                    {
                        "featureName": "Sepal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.225,
                        "impactNormalized": 1.0,
                    },
                    {
                        "featureName": "Sepal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.057,
                        "impactNormalized": 0.257,
                    },
                ],
                "class": "setosa",
            },
            {
                "featureImpacts": [
                    {
                        "featureName": "Petal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.302,
                        "impactNormalized": 1.0,
                    },
                    {
                        "featureName": "Petal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.229,
                        "impactNormalized": 0.758,
                    },
                    {
                        "featureName": "Sepal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.221,
                        "impactNormalized": 0.730,
                    },
                    {
                        "featureName": "Sepal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.073,
                        "impactNormalized": 0.241,
                    },
                ],
                "class": "versicolor",
            },
            {
                "featureImpacts": [
                    {
                        "featureName": "Petal Width",
                        "redundantWith": None,
                        "impactUnnormalized": 0.273,
                        "impactNormalized": 1.0,
                    },
                    {
                        "featureName": "Petal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.174,
                        "impactNormalized": 0.640,
                    },
                    {
                        "featureName": "Sepal Width",
                        "redundantWith": None,
                        "impactUnnormalized": -0.001,
                        "impactNormalized": -0.006,
                    },
                    {
                        "featureName": "Sepal Length",
                        "redundantWith": None,
                        "impactUnnormalized": 0.018,
                        "impactNormalized": 0.067,
                    },
                ],
                "class": "virginica",
            },
        ],
    }


@pytest.fixture
def feature_impact_response(feature_impact_server_data, feature_impact_url):
    body = json.dumps(feature_impact_server_data)
    responses.add(
        responses.GET, feature_impact_url, status=200, content_type="application/json", body=body
    )


@pytest.fixture
def multiclass_feature_impact_response(
    multiclass_feature_impact_server_data, multiclass_feature_impact_url
):
    body = json.dumps(multiclass_feature_impact_server_data)
    responses.add(
        responses.GET,
        multiclass_feature_impact_url,
        status=200,
        content_type="application/json",
        body=body,
    )


@pytest.fixture
def feature_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType="featureImpact")


@pytest.fixture
def feature_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType="featureImpact")


@pytest.fixture
def approximate_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def approximate_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def prime_validation_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def prime_validation_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def feature_impact_completed_response(
    feature_impact_job_finished_server_data, job_url, feature_impact_url
):
    """
    Loads a response that the given job is a featureImpact job, and is in completed
    """
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_impact_job_finished_server_data),
        status=303,
        adding_headers={"Location": feature_impact_url},
        content_type="application/json",
    )


@pytest.fixture
def feature_impact_previously_ran_response(project_url, model_id, job_id):
    """
    Loads a response that the given model has already ran its feature impact
     """
    body = {
        "message": "Feature Impact is in progress for this model.",
        "errorName": "JobAlreadyAdded",
        "jobId": job_id,
    }
    responses.add(
        responses.POST,
        "{}models/{}/featureImpact/".format(project_url, model_id),
        body=json.dumps(body),
        status=422,
        content_type="application/json",
    )


@pytest.fixture
def feature_impact_running_response(feature_impact_job_running_server_data, job_url):
    """
    Loads a response that the given job is a featureImpact job, and is running
    """
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(feature_impact_job_running_server_data),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def approximate_job_creation_response(project_url, model_id, job_url):
    responses.add(
        responses.POST,
        "{}models/{}/primeRulesets/".format(project_url, model_id),
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def approximate_completed_response(
    approximate_job_finished_server_data,
    job_url,
    project_url,
    model_id,
    ruleset_without_model_server_data,
    ruleset_with_model_server_data,
):
    rulesets_url = "{}models/{}/primeRulesets/".format(project_url, model_id)
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(approximate_job_finished_server_data),
        status=303,
        adding_headers={"Location": rulesets_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        rulesets_url,
        body=json.dumps([ruleset_with_model_server_data, ruleset_without_model_server_data]),
        content_type="application/json",
    )


@pytest.fixture
def prime_validation_job_creation_response(project_url, job_url):
    responses.add(
        responses.POST,
        "{}primeFiles/".format(project_url),
        body="",
        status=202,
        adding_headers={"Location": job_url},
    )


@pytest.fixture
def prime_validation_job_completed_response(
    prime_validation_job_finished_server_data, job_url, project_url, prime_file_server_data
):
    file_url = "{}primeFiles/{}/".format(project_url, prime_file_server_data["id"])
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(prime_validation_job_finished_server_data),
        status=303,
        adding_headers={"Location": file_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        file_url,
        body=json.dumps(prime_file_server_data),
        content_type="application/json",
    )


@pytest.fixture
def mocked_execution_environment_versions():
    return {
        "count": 5,
        "totalCount": 5,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "buildStatus": "success",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a2a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "",
                "buildStatus": "processing",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
                "dockerContextSize": 1024,
                "dockerImageSize": 2048,
            },
            {
                "id": "5cf4d3f5f930e26daac18a3a",
                "environmentId": "5cf4d3f5f930e26daac18a1xx",
                "label": "Version 1",
                "description": "some description",
                "buildStatus": "failed",
                "created": "2019-09-28T15:19:26.587583Z",
                "dockerContextSize": None,
                "dockerImageSize": None,
            },
        ],
    }


@pytest.fixture
def mocked_execution_environment_version(mocked_execution_environment_versions):
    return mocked_execution_environment_versions["data"][0]


@pytest.fixture
def mocked_execution_environments(mocked_execution_environment_version):
    return {
        "count": 2,
        "totalCount": 2,
        "next": None,
        "previous": None,
        "data": [
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env1",
                "description": "some description",
                "programmingLanguage": "python",
                "isPublic": True,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": mocked_execution_environment_version,
            },
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env2",
                "description": "",
                "programmingLanguage": "other",
                "isPublic": False,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": None,
            },
            {
                "id": "5cf4d3f5f930e26daac18a1a",
                "name": "env2",
                "description": "",
                "programmingLanguage": "other",
                "isPublic": False,
                "created": "2019-09-28T15:19:26.587583Z",
                "latestVersion": None,
                "requiredMetadataKeys": [
                    {"fieldName": "FIELD_{}".format(i), "displayName": "field {}".format(i)}
                    for i in range(3)
                ],
            },
        ],
    }


@pytest.fixture
def mocked_execution_environment(mocked_execution_environments):
    return mocked_execution_environments["data"][0]


@pytest.fixture
def mocked_execution_environment_with_required_metadata_keys(mocked_execution_environments):
    return next(
        env for env in mocked_execution_environments["data"] if env.get("requiredMetadataKeys")
    )


@pytest.fixture
def mocked_custom_model_version_with_required_metadata(mocked_versions):
    return next(version for version in mocked_versions["data"] if version.get("requiredMetadata"))


@pytest.fixture
def shap_impact_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/shapImpact/".format(project_id, model_id)


@pytest.fixture
def shap_impact_job_creation_response(shap_impact_url, job_url):
    responses.add(
        responses.POST, shap_impact_url, body="", status=202, adding_headers={"Location": job_url}
    )


@pytest.fixture
def shap_impact_response(shap_impact_server_data, shap_impact_url):
    body = json.dumps(shap_impact_server_data)
    responses.add(
        responses.GET, shap_impact_url, status=200, content_type="application/json", body=body
    )


@pytest.fixture
def shap_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType="shapImpact")


@pytest.fixture
def shap_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType="shapImpact")


@pytest.fixture
def shap_impact_running_response(shap_impact_job_running_server_data, job_url):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(shap_impact_job_running_server_data),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def shap_impact_completed_response(shap_impact_job_finished_server_data, job_url, shap_impact_url):
    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(shap_impact_job_finished_server_data),
        status=303,
        adding_headers={"Location": shap_impact_url},
        content_type="application/json",
    )


@pytest.fixture
def shap_impact_previously_ran_response(project_url, model_id, job_id, job_url):
    body = {
        "message": "SHAP impact already exists.",
        "errorName": "JobAlreadyAdded",
        "jobId": job_id,
    }
    responses.add(
        responses.POST,
        "{}models/{}/shapImpact/".format(project_url, model_id),
        body=json.dumps(body),
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )



# -*- coding: utf-8 -*-
from copy import deepcopy
import json

import numpy as np
import pandas as pd
import pytest
import responses
import trafaret

from datarobot import errors
from datarobot.models.shap_matrix import ShapMatrix
from tests.utils import add_response, request_body_to_json


@pytest.fixture
def shap_matrix_id():
    return "5ea6e6ab4cfad16f3499940e"


@pytest.fixture
def mocked_shap_matrix_values():
    row_indexes = [u"row_{}".format(i) for i in range(10)]
    column_names = [u"カテゴリー_{}".format(i) for i in range(5)]
    rs = np.random.RandomState(1234)
    shap_scores = rs.uniform(-1, 1, (len(row_indexes), len(column_names)))
    df = pd.DataFrame(data=shap_scores, columns=column_names, index=row_indexes)
    return df.to_csv(encoding="utf-8")


@pytest.fixture
def mocked_shap_matrices(shap_matrix_id, project_id, model_id, dataset_id):
    shap_matrices_json = [
        {
            "id": shap_matrix_id,
            "projectId": project_id,
            "modelId": model_id,
            "datasetId": dataset_id,
        },
    ]
    return {
        "count": len(shap_matrices_json),
        "next": None,
        "previous": None,
        "data": shap_matrices_json,
    }


@pytest.fixture
def mocked_shap_matrix(mocked_shap_matrices):
    return mocked_shap_matrices["data"][0]


@pytest.fixture
def make_shap_matrices_url(unittest_endpoint, project_id):
    def _make_shap_matrices_url(shap_matrix_id=None):
        base_url = unittest_endpoint + "/projects/{}/shapMatrices/".format(project_id)
        if shap_matrix_id is not None:
            return "{}{}/".format(base_url, shap_matrix_id)
        return base_url

    return _make_shap_matrices_url


def assert_shap_matrix(shap_matrix, shap_matrix_json):
    assert shap_matrix.id == shap_matrix_json["id"]
    assert shap_matrix.project_id == shap_matrix_json["projectId"]
    assert shap_matrix.model_id == shap_matrix_json["modelId"]
    assert shap_matrix.dataset_id == shap_matrix_json["datasetId"]


def test_from_server_data(mocked_shap_matrix):
    shap_matrix = ShapMatrix.from_server_data(mocked_shap_matrix)
    assert_shap_matrix(shap_matrix, mocked_shap_matrix)


@responses.activate
def test_get__no_request_made(project_id, shap_matrix_id):
    ShapMatrix.get(project_id, shap_matrix_id)
    assert len(responses.calls) == 0


@responses.activate
def test_list(project_id, mocked_shap_matrices, make_shap_matrices_url):
    url = make_shap_matrices_url()
    add_response(url, mocked_shap_matrices)

    shap_matrices = ShapMatrix.list(project_id)
    assert shap_matrices is not None
    assert len(mocked_shap_matrices["data"]) == len(shap_matrices)
    for shap_matrix, shap_matrix_json in zip(shap_matrices, mocked_shap_matrices["data"]):
        assert_shap_matrix(shap_matrix, shap_matrix_json)
    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url == url


@responses.activate
def test_list_multiple_pages(project_id, mocked_shap_matrices, make_shap_matrices_url):
    url1 = make_shap_matrices_url()
    url2 = make_shap_matrices_url() + "2"

    mocked_shap_matrices_2nd = deepcopy(mocked_shap_matrices)
    mocked_shap_matrices["next"] = url2

    add_response(url1, mocked_shap_matrices)
    add_response(url2, mocked_shap_matrices_2nd)

    shap_matrices = ShapMatrix.list(project_id)
    assert len(shap_matrices) == len(mocked_shap_matrices["data"]) + len(
        mocked_shap_matrices_2nd["data"]
    )
    for shap_matrix, shap_matrix_json in zip(
        shap_matrices, mocked_shap_matrices["data"] + mocked_shap_matrices_2nd["data"]
    ):
        assert_shap_matrix(shap_matrix, shap_matrix_json)

    assert responses.calls[0].request.method == "GET"
    assert responses.calls[0].request.url.endswith(url1)
    assert responses.calls[1].request.method == "GET"
    assert responses.calls[1].request.url.endswith(url2)


@responses.activate
def test_list__no_id_in_api_response_causes_error(
    project_id, mocked_shap_matrices, make_shap_matrices_url
):
    mocked_shap_matrices = deepcopy(mocked_shap_matrices)
    shap_matrix_no_id = {
        "id": None,
        "projectId": project_id,
        "modelId": "model-id",
        "datasetId": "dataset-id",
    }
    mocked_shap_matrices["data"].append(shap_matrix_no_id)

    url = make_shap_matrices_url()
    add_response(url, mocked_shap_matrices)

    with pytest.raises(trafaret.DataError):
        ShapMatrix.list(project_id)


@responses.activate
def test_get_as_dataframe(
    project_id, shap_matrix_id, mocked_shap_matrix_values, make_shap_matrices_url
):
    url = make_shap_matrices_url(shap_matrix_id)
    add_response(url, mocked_shap_matrix_values, content_type="text/csv")

    shap_matrix = ShapMatrix.get(project_id, shap_matrix_id)
    shap_matrix_values = shap_matrix.get_as_dataframe()
    assert isinstance(shap_matrix_values, pd.DataFrame)
    assert shap_matrix_values.shape == (10, 5)


@responses.activate
def test_get_as_dataframe__not_found(
    project_id, shap_matrix_id, mocked_shap_matrix_values, make_shap_matrices_url
):
    url = make_shap_matrices_url(shap_matrix_id)
    add_response(url, "", status=404)

    shap_matrix = ShapMatrix.get(project_id, shap_matrix_id)
    with pytest.raises(errors.ClientError):
        shap_matrix.get_as_dataframe()


@responses.activate
def test_get_as_dataframe__unknown_error(
    project_id, shap_matrix_id, mocked_shap_matrix_values, make_shap_matrices_url
):
    url = make_shap_matrices_url(shap_matrix_id)
    add_response(url, "", status=500)

    shap_matrix = ShapMatrix.get(project_id, shap_matrix_id)
    with pytest.raises(errors.ServerError):
        shap_matrix.get_as_dataframe()


@responses.activate
def test_create(one_model, project_url, base_job_completed_server_data):
    shap_matrices_url = project_url + "shapMatrices/"
    job_data = dict(base_job_completed_server_data, jobType="shapMatrix")
    job_id = job_data["id"]
    job_url = project_url + "jobs/{}/".format(job_id)
    dataset_id = "12344321beefcafe43211234"
    shap_matrix_id = "deadbeef12344321feebdaed"
    finished_shap_matrix_url = shap_matrices_url + "{}/".format(shap_matrix_id)

    submit_shap_job_request = {"modelId": one_model.id, "datasetId": dataset_id}
    responses.add(
        responses.POST,
        shap_matrices_url,
        body=json.dumps(submit_shap_job_request),
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )

    responses.add(
        responses.GET,
        job_url,
        body=json.dumps(job_data),
        status=303,
        content_type="application/json",
        adding_headers={"Location": finished_shap_matrix_url},
    )

    shap_matrix_job = ShapMatrix.create(
        project_id=one_model.project_id, model_id=one_model.id, dataset_id=dataset_id
    )
    assert shap_matrix_job.id == int(job_id)

    request_json = request_body_to_json(responses.calls[0].request)
    assert request_json.get("modelId") == submit_shap_job_request["modelId"]
    assert request_json.get("datasetId") == submit_shap_job_request["datasetId"]


@responses.activate
def test_create__client_error(project_url, one_model, dataset_id):
    error_message = {"message": "some error message"}
    responses.add(
        responses.POST,
        project_url + "shapMatrices/",
        body=json.dumps(error_message),
        status=422,
        content_type="application/json",
    )
    with pytest.raises(errors.ClientError):
        ShapMatrix.create(
            project_id=one_model.project_id, model_id=one_model.id, dataset_id=dataset_id
        )


@responses.activate
def test_create__server_error(project_url, one_model, dataset_id):
    shap_matrices_url = project_url + "shapMatrices/"
    submit_shap_job_request = {"modelId": one_model.id, "datasetId": dataset_id}
    responses.add(
        responses.POST,
        shap_matrices_url,
        body=json.dumps(submit_shap_job_request),
        status=500,
        content_type="application/json",
    )
    with pytest.raises(errors.ServerError):
        ShapMatrix.create(
            project_id=one_model.project_id, model_id=one_model.id, dataset_id=dataset_id
        )



import trafaret as t

from datarobot.errors import ClientError
from datarobot.models.roc_curve import RocCurveThresholdMixin, RocPointsTrafaret
from datarobot.utils.pagination import unpaginate

from ...utils import encode_utf8_if_py2
from ..api_object import APIObject
from .external_scores import DEFAULT_BATCH_SIZE


class ExternalRocCurve(APIObject, RocCurveThresholdMixin):
    """ ROC curve data for the model and prediction dataset with target or actual value column in
    unsupervised case.

    .. versionadded:: v2.21

    Attributes
    ----------
    dataset_id: str
        id of the prediction dataset with target or actual value column for unsupervised case
    roc_points: list of dict
        List of precalculated metrics associated with thresholds for ROC curve.
    negative_class_predictions: list of float
        List of predictions from example for negative class
    positive_class_predictions: list of float
        List of predictions from example for positive class
    """

    _path = "projects/{project_id}/models/{model_id}/datasetRocCurves/"

    _converter = t.Dict({t.Key("dataset_id"): t.String}).merge(RocPointsTrafaret).ignore_extra("*")

    def __init__(
        self, dataset_id, roc_points, negative_class_predictions, positive_class_predictions
    ):
        self.dataset_id = dataset_id
        self.roc_points = roc_points
        self.negative_class_predictions = negative_class_predictions
        self.positive_class_predictions = positive_class_predictions

    def __repr__(self):
        return encode_utf8_if_py2(
            u"ExternalRocCurve(dataset_id={}, roc_points={})".format(
                self.dataset_id, self.roc_points
            )
        )

    @classmethod
    def list(cls, project_id, model_id, dataset_id=None, offset=0, limit=100):
        """ Retrieve list of the roc curves for the model.

        Parameters
        ----------
        project_id: str
            id of the project
        model_id: str
            if specified, only lift chart for this model will be retrieved
        dataset_id: str, optional
            if specified, only lift chart for this dataset will be retrieved
        offset: int, optional
            this many results will be skipped, default: 0
        limit: int, optional
            at most this many results are returned, default: 100, max 1000.
            To return all results, specify 0

        Returns
        -------
            A list of :py:class:`ExternalRocCurve <datarobot.ExternalRocCurve>` objects
        """
        url = cls._path.format(project_id=project_id, model_id=model_id)
        params = {"limit": limit, "offset": offset}
        if dataset_id:
            params["datasetId"] = dataset_id
        if limit == 0:  # unlimited results
            params["limit"] = DEFAULT_BATCH_SIZE
            return [cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)]
        r_data = cls._client.get(url, params=params).json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    @classmethod
    def get(cls, project_id, model_id, dataset_id):
        """ Retrieve ROC curve chart for the model and prediction dataset.

        Parameters
        ----------
        project_id: str
            project id
        model_id: str
            model id
        dataset_id: str
            prediction dataset id with target or actual value column for unsupervised case

        Returns
        -------
            :py:class:`ExternalRocCurve <datarobot.ExternalRocCurve>` object

        """
        if dataset_id is None:
            raise ValueError("dataset_id must be specified")
        charts = cls.list(project_id, model_id, dataset_id=dataset_id)
        if not charts:
            raise ClientError("Requested roc curve does not exist.", 404)
        return charts[0]



import responses

from datarobot import RelationshipsConfiguration


@responses.activate
def test_relationships_config_creation(
    relationships_configuration, create_dataset_definitions, create_relationships
):
    responses.add(
        responses.POST,
        "https://host_name.com/relationshipsConfigurations/",
        json=relationships_configuration,
    )

    relationships_configuration = RelationshipsConfiguration.create(
        dataset_definitions=create_dataset_definitions, relationships=create_relationships,
    )
    assert responses.calls[0].request.method == "POST"
    assert relationships_configuration.id == "5a530498d5c1f302d6d176c8"
    assert len(relationships_configuration.dataset_definitions) == 2
    assert len(relationships_configuration.relationships) == 2


@responses.activate
def test_relationships_config_retrieve(
    relationships_configuration, create_dataset_definitions, create_relationships
):
    id = relationships_configuration["id"]
    responses.add(
        responses.GET,
        "https://host_name.com/relationshipsConfigurations/{}/".format(id),
        json=relationships_configuration,
    )

    result = RelationshipsConfiguration(id=id).get()

    assert responses.calls[0].request.method == "GET"
    assert result.id == id
    assert len(result.dataset_definitions) == 2
    assert len(result.relationships) == 2
    for actual, expected in zip(result.dataset_definitions, create_dataset_definitions):
        assert actual["identifier"] == expected["identifier"]
        assert actual["catalog_version_id"] == expected["catalogVersionId"]
        assert actual["catalog_id"] == expected["catalogId"]
        assert actual["snapshot_policy"] == expected["snapshotPolicy"]

    actual = result.relationships[1]
    expected = create_relationships[1]
    assert actual["dataset1_identifier"] == expected["dataset1Identifier"]
    assert actual["dataset2_identifier"] == expected["dataset2Identifier"]
    assert actual["dataset1_keys"] == expected["dataset1Keys"]
    assert actual["dataset2_keys"] == expected["dataset2Keys"]
    assert actual["feature_derivation_window_end"] == expected["featureDerivationWindowEnd"]
    assert actual["feature_derivation_window_start"] == expected["featureDerivationWindowStart"]
    assert (
        actual["feature_derivation_window_time_unit"] == expected["featureDerivationWindowTimeUnit"]
    )
    assert actual["prediction_point_rounding"] == expected["predictionPointRounding"]
    assert (
        actual["prediction_point_rounding_time_unit"] == expected["predictionPointRoundingTimeUnit"]
    )


@responses.activate
def test_relationships_config_replace(
    relationships_configuration, create_dataset_definitions, create_relationships
):
    responses.add(
        responses.PUT,
        "https://host_name.com/relationshipsConfigurations/R-ID/",
        json=relationships_configuration,
    )

    relationships_configuration = RelationshipsConfiguration("R-ID").replace(
        dataset_definitions=create_dataset_definitions, relationships=create_relationships,
    )
    assert responses.calls[0].request.method == "PUT"
    assert relationships_configuration.id == "5a530498d5c1f302d6d176c8"
    assert len(relationships_configuration.dataset_definitions) == 2
    assert len(relationships_configuration.relationships) == 2


@responses.activate
def test_relationships_config_delete(relationships_configuration):
    id = relationships_configuration["id"]
    url = "https://host_name.com/relationshipsConfigurations/{}/".format(id)
    responses.add(responses.DELETE, url)

    relationships_config = RelationshipsConfiguration(id=id)
    relationships_config.delete()

    assert responses.calls[0].request.method == responses.DELETE
    assert responses.calls[0].request.url == url



import trafaret as t

from ..api_object import APIObject
from .images import Image

__all__ = ["ImageEmbedding", "ImageActivationMap"]


class ImageEmbedding(APIObject):
    """Vector representation of an image in an embedding space.

    A vector in an embedding space will allow linear computations to
    be carried out between images: for example computing the Euclidean
    distance of the images.

    Attributes
    ----------
    image: Image
        Image object used to create this map.
    feature_name: str
        Name of the feature column this embedding is associated with.
    position_x: int
        X coordinate of the image in the embedding space.
    position_y: int
        Y coordinate of the image in the embedding space.
    actual_target_value: object
        Actual target value of the dataset row.
    """

    _compute_path = "projects/{project_id}/models/{model_id}/imageEmbeddings/"
    _models_path = "projects/{project_id}/imageEmbeddings/"
    _list_path = "projects/{project_id}/models/{model_id}/imageEmbeddings/"
    _converter = t.Dict(
        {
            t.Key("image_id", optional=True): t.String(),
            t.Key("position_x", optional=True): t.Float(),
            t.Key("position_y", optional=True): t.Float(),
            t.Key("actual_target_value", optional=True): t.Any(),
            t.Key("target_values", optional=True): t.Or(t.List(t.String()), t.Null),
            t.Key("target_bins", optional=True): t.Or(t.List(t.Any()), t.Null),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self.project_id = kwargs.get("project_id")
        self.model_id = kwargs.get("model_id")
        self.image = Image(**kwargs)
        self.feature_name = kwargs.get("feature_name")
        self.position_x = kwargs.get("position_x")
        self.position_y = kwargs.get("position_y")
        self.actual_target_value = kwargs.get("actual_target_value")

    def __repr__(self):
        return (
            "datarobot.models.visualai.ImageEmbedding("
            "project_id={0.project_id}, "
            "model_id={0.model_id}, "
            "feature_name={0.feature_name}, "
            "position_x={0.position_x}, "
            "position_y={0.position_y}, "
            "image_id={0.image.id})"
        ).format(self)

    @classmethod
    def compute(cls, project_id, model_id):
        """Start creation of image embeddings for the model.

        Parameters
        ----------
        project_id: str
            Project to start creation in.
        model_id: str
            Project's model to start creation in.

        Returns
        -------
        str
            URL to check for image embeddings progress.

        Raises
        ------
        datarobot.errors.ClientError
            Server rejected creation due to client error. Most likely
            cause is bad ``project_id`` or ``model_id``.
        """
        path = cls._compute_path.format(project_id=project_id, model_id=model_id)
        r_data = cls._client.post(path).json()
        return r_data["url"]

    @classmethod
    def models(cls, project_id):
        """List the models in a project.

        Parameters
        ----------
        project_id: str
            Project that contains the models.

        Returns
        -------
        list( tuple(model_id, feature_name) )
             List of model and feature name pairs.
        """
        path = cls._models_path.format(project_id=project_id)
        r_data = cls._client.get(path).json()
        return [(d["modelId"], d["featureName"]) for d in r_data.get("data", [])]

    @classmethod
    def list(cls, project_id, model_id, feature_name):
        """Return a list of ImageEmbedding objects.

        Parameters
        ----------
        project_id: str
            Project that contains the images.
        model_id: str
            Model that contains the images.
        feature_name: str
            Name of feature column that contains images.
        """
        path = cls._list_path.format(project_id=project_id, model_id=model_id)
        list_params = {}
        list_params["featureName"] = feature_name
        r_data = cls._client.get(path, params=list_params).json()
        ret = []
        for embed_data in r_data.get("embeddings", []):
            embed = cls.from_server_data(embed_data)
            embed.project_id = project_id
            embed.model_id = model_id
            embed.feature_name = feature_name
            embed.image.project_id = project_id
            ret.append(embed)
        return ret


class ImageActivationMap(APIObject):
    """Mark areas of image with weight of impact on training.

    This is a technique to display how various areas of the region were
    used in training, and their effect on predictions. Larger values in
    ``activation_values`` indicates a larger impact.

    Attributes
    ----------
    image: Image
        Image object used to create this map.
    overlay_image: Image
        Image object composited with activation heat map.
    feature_name: str
        Name of the feature column that contains the value this map is
        based on.
    height: int
        Height of the original image in pixels.
    width: int
        Width of the original image in pixels.
    actual_target_value: object
        Actual target value of the dataset row.
    predicted_target_value: object
        Predicted target value of the dataset row that contains this image.
    activation_values: [ [ int ] ]
        A row-column matrix that contains the activation strengths for
        image regions. Values are integers in the range [0, 255].
    """

    _compute_path = "projects/{project_id}/models/{model_id}/imageActivationMaps/"
    _models_path = "projects/{project_id}/imageActivationMaps/"
    _list_path = "projects/{project_id}/models/{model_id}/imageActivationMaps/"
    _converter = t.Dict(
        {
            t.Key("image_id", optional=True): t.String(),
            t.Key("overlay_image_id", optional=True): t.String(),
            t.Key("feature_name", optional=True): t.String(),
            t.Key("image_width", to_name="width", optional=True): t.Int(),
            t.Key("image_height", to_name="height", optional=True): t.Int(),
            t.Key("activation_values", optional=True): t.List(t.List(t.Int())),
            t.Key("actual_target_value", optional=True): t.Any(),
            t.Key("predicted_target_value", optional=True): t.Any(),
            t.Key("target_values", optional=True): t.Or(t.List(t.String()), t.Null),
            t.Key("target_bins", optional=True): t.Or(t.List(t.Any()), t.Null),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs):
        self.project_id = kwargs.get("project_id")
        self.image = Image(**kwargs)
        self.overlay_image = Image(
            image_id=kwargs.get("overlay_image_id", kwargs.get("image_id")),
            project_id=kwargs.get("project_id"),
            height=kwargs.get("height", 0),
            width=kwargs.get("width", 0),
        )
        self.feature_name = kwargs.get("feature_name")
        self.actual_target_value = kwargs.get("actual_target_value")
        self.predicted_target_value = kwargs.get("predicted_target_value")
        self.activation_values = kwargs.get("activation_values")

    def __repr__(self):
        return (
            "datarobot.models.visualai.ActivationMap("
            "project_id={0.project_id}, "
            "model_id={0.model_id}, "
            "feature_name={0.feature_name}, "
            "image_id={0.image.id}, "
            "overlay_image_id={0.overlay_image.id}, "
            "height={0.image.height}, "
            "width={0.image.width})"
        ).format(self)

    @classmethod
    def compute(cls, project_id, model_id):
        """Start creation of a activation map in the given model.

        Parameters
        ----------
        project_id: str
            Project to start creation in.
        model_id: str
            Project's model to start creation in.

        Returns
        -------
        str
            URL to check for image embeddings progress.

        Raises
        ------
        datarobot.errors.ClientError
            Server rejected creation due to client error. Most likely
            cause is bad ``project_id`` or ``model_id``.
        """
        path = cls._compute_path.format(project_id=project_id, model_id=model_id)
        r_data = cls._client.post(path).json()
        return r_data["url"]

    @classmethod
    def models(cls, project_id):
        """List the models in a project.

        Parameters
        ----------
        project_id: str
            Project that contains the models.

        Returns
        -------
        list( tuple(model_id, feature_name) )
             List of model and feature name pairs.
        """
        path = cls._models_path.format(project_id=project_id)
        r_data = cls._client.get(path).json()
        return [(d["modelId"], d["featureName"]) for d in r_data.get("data", [])]

    @classmethod
    def list(cls, project_id, model_id, feature_name, offset=None, limit=None):
        """Return a list of ImageActivationMap objects.

        Parameters
        ----------
        project_id: str
            Project that contains the images.
        model_id: str
            Model that contains the images.
        feature_name: str
            Name of feature column that contains images.
        offset: int
            Number of images to be skipped.
        limit: int
            Number of images to be returned.
        """
        path = cls._list_path.format(project_id=project_id, model_id=model_id)
        list_params = {}
        list_params["featureName"] = feature_name
        if offset:
            list_params["offset"] = int(offset)
        if limit:
            list_params["limit"] = int(limit)
        r_data = cls._client.get(path, params=list_params).json()
        ret = []
        for amap_data in r_data.get("activationMaps", []):
            amap = cls.from_server_data(amap_data)
            amap.project_id = project_id
            amap.model_id = model_id
            amap.image.project_id = project_id
            amap.overlay_image.project_id = project_id
            ret.append(amap)
        return ret



from copy import deepcopy
import json

import pytest
import responses

from datarobot import (
    enums,
    FeatureImpactJob,
    Job,
    Project,
    TrainingPredictions,
    TrainingPredictionsJob,
)
from datarobot.errors import AsyncProcessUnsuccessfulError, AsyncTimeoutError

mock_job_data = {
    "status": "inprogress",
    "url": "https://host_name.com/projects/p-id/modelJobs/1/",
    "id": "1",
    "jobType": "model",
    "isBlocked": False,
    "modelId": "5d518cd3962d741512605e2b",
    "projectId": "p-id",
}

mock_model_job_data = {
    "status": enums.QUEUE_STATUS.INPROGRESS,
    "processes": [
        "One-Hot Encoding",
        "Bernoulli Naive Bayes classifier (scikit-learn)",
        "Missing Values Imputed",
        "Gaussian Naive Bayes classifier (scikit-learn)",
        "Naive Bayes combiner classifier",
        "Calibrate predictions",
    ],
    "projectId": "p-id",
    "samplePct": 28.349,
    "modelType": "Naive Bayes combiner classifier",
    "featurelistId": "56d8620bccf94e26f37af0a3",
    "modelCategory": "model",
    "blueprintId": "2a1b9ae97fe61880332e196c770c1f9f",
    "id": "1",
}

SAMPLE_BINARY_PREDICTION_RETURN = json.dumps(
    {
        u"positiveClass": 1.0,
        u"task": u"Binary",
        u"predictions": [
            {
                u"positiveProbability": 0.9,
                u"prediction": 1.0,
                u"rowId": 0,
                u"predictionValues": [
                    {u"value": 0.9, u"label": u"1.0"},
                    {u"value": 0.1, u"label": u"0.0"},
                ],
            },
            {
                u"positiveProbability": 0.1,
                u"prediction": 0.0,
                u"rowId": 1,
                u"predictionValues": [
                    {u"value": 0.9, u"label": u"1.0"},
                    {u"value": 0.1, u"label": u"0.0"},
                ],
            },
            {
                u"positiveProbability": 0.8,
                u"prediction": 1.0,
                u"rowId": 2,
                u"predictionValues": [
                    {u"value": 0.8, u"label": u"1.0"},
                    {u"value": 0.2, u"label": u"0.0"},
                ],
            },
        ],
    }
)


def assert_job_attributes(job):
    assert job.id == 1
    assert job.project == Project("p-id")
    assert job.status == "inprogress"
    assert job.job_type == "model"
    assert job.model_id == "5d518cd3962d741512605e2b"
    # ensure the instantiated object has the model ID as well
    assert job.model.id == "5d518cd3962d741512605e2b"
    assert not job.is_blocked
    assert job._job_details_path == "projects/p-id/modelJobs/1/"


def test_instantiate():
    new_job = Job(mock_job_data)
    assert_job_attributes(new_job)


def test_future_proof():
    Job(dict(mock_job_data, future="new"))


@responses.activate
@pytest.mark.parametrize("job_class", (Job, FeatureImpactJob))
def test_get_completed(client, job_class):
    resource_url = "https://host_name.com/resource-location/"
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(dict(mock_job_data, status="COMPLETED")),
        content_type="application/json",
        adding_headers={"Location": resource_url},
    )
    job = job_class.get("p-id", "1")
    assert job.status == "COMPLETED"
    assert job._completed_resource_url == resource_url


@responses.activate
def test_get_completed_feature_impact(client, feature_impact_server_data):
    job_result_url = "https://host_name.com/resource-location/"
    pid = "p-id"

    # Get project.
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/".format(pid),
        status=200,
        body=json.dumps({"id": pid}),
        content_type="application/json",
    )
    project = Project.get(pid)

    # Get not finished jobs.
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/jobs/".format(pid),
        status=200,
        body=json.dumps({"jobs": [mock_job_data]}),
        content_type="application/json",
    )
    job = project.get_all_jobs()[0]

    # Mark job as finished.
    job_data = deepcopy(mock_job_data)
    job_data["status"] = "COMPLETED"
    job_data["jobType"] = "featureImpact"

    # Get that finished job and its results.
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/jobs/1/".format(pid),
        status=303,
        body=json.dumps(job_data),
        content_type="application/json",
        adding_headers={"Location": job_result_url},
    )
    responses.add(
        responses.GET,
        job_result_url,
        status=200,
        body=json.dumps(feature_impact_server_data),
        content_type="application/json",
    )
    assert job.get_result() == feature_impact_server_data["featureImpacts"]


@responses.activate
@pytest.mark.parametrize("job_class", (Job, FeatureImpactJob))
def test_get_no_model_id(client, job_class):
    resource_url = "https://host_name.com/resource-location/"
    job_data = mock_job_data.copy()
    del job_data["modelId"]
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(job_data),
        content_type="application/json",
        adding_headers={"Location": resource_url},
    )
    job = job_class.get("p-id", "1")
    assert job.model_id is None
    assert job.model is None


@responses.activate
def test_get_completed_pred(client):
    get_predictions_url = "https://host_name.com/projects/p-id/predictions/1/"

    mock_generic_job_data = {
        "status": "COMPLETED",
        "url": "https://host_name.com/projects/p-id/predictJobs/1/",
        "id": "1",
        "jobType": "predict",
        "isBlocked": False,
        "projectId": "p-id",
    }

    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(dict(mock_generic_job_data, status="COMPLETED")),
        content_type="application/json",
        adding_headers={"Location": get_predictions_url},
    )
    responses.add(
        responses.GET,
        get_predictions_url,
        status=200,
        body=SAMPLE_BINARY_PREDICTION_RETURN,
        content_type="application/json",
    )

    job = Job.get("p-id", "1")

    data = job.get_result()
    assert job.status == "COMPLETED"

    # inspect the data
    expected_columns = ["positive_probability", "prediction", "row_id", "class_0.0", "class_1.0"]
    assert sorted(data.columns.values) == sorted(expected_columns)


@responses.activate
def test_wait_for_completion_never_finishes(client, mock_async_time):
    resource_url = "https://host_name.com/resource-location/"
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=200,
        body=json.dumps(mock_job_data),
        content_type="application/json",
        adding_headers={"Location": resource_url},
    )
    job = Job(mock_job_data)
    mock_async_time.time.side_effect = (0, 2)
    with pytest.raises(AsyncTimeoutError):
        job.wait_for_completion(max_wait=1)


@responses.activate
@pytest.mark.parametrize("job_class", (Job, FeatureImpactJob))
def test_wait_for_completion_finishes(client, mock_async_time, job_class):
    resource_url = "https://host_name.com/resource-location/"
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(dict(mock_job_data, status="COMPLETED")),
        content_type="application/json",
        adding_headers={"Location": resource_url},
    )
    mock_async_time.time.side_effect = (0, 2)
    job = job_class(mock_job_data)
    job.wait_for_completion(max_wait=3)
    assert job.status == "COMPLETED"


@responses.activate
def test_get_result_errored(project_id, project_url):
    job_url = "{}jobs/1/".format(project_url)
    mock_job = {
        "status": enums.QUEUE_STATUS.ERROR,
        "url": "{}modelJobs/1/".format(project_url),
        "id": "1",
        "jobType": "model",
        "isBlocked": False,
        "projectId": project_id,
    }
    responses.add(
        responses.GET, job_url, status=200, json=mock_job, content_type="application/json"
    )
    job = Job.get(project_id, "1")
    with pytest.raises(AsyncProcessUnsuccessfulError):
        job.get_result()


@responses.activate
@pytest.mark.parametrize("job_class", (Job, FeatureImpactJob))
def test_get(client, job_class):
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=200,
        body=json.dumps(mock_job_data),
        content_type="application/json",
    )
    job = job_class.get("p-id", "1")
    assert_job_attributes(job)
    assert job._completed_resource_url is None


@responses.activate
def test_refresh(client):
    job = Job(mock_job_data)
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=200,
        body=json.dumps(refresh_job_data),
        content_type="application/json",
    )
    job.refresh()
    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_cancel_job(client):
    responses.add(
        responses.DELETE,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=204,
        body="",
        content_type="application/json",
    )
    generic_job = Job(mock_job_data)
    generic_job.cancel()


@responses.activate
def test_training_predictions_job__refresh__ok():
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.ERROR)
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=200,
        body=json.dumps(refresh_job_data),
        content_type="application/json",
    )

    job = TrainingPredictionsJob(mock_job_data, model_id="model-id", data_subset="all")
    job.refresh()

    assert job.status == enums.QUEUE_STATUS.ERROR


@responses.activate
def test_training_predictions_job__get__ok():
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=200,
        body=json.dumps(mock_job_data),
        content_type="application/json",
    )

    job = TrainingPredictionsJob.get("p-id", "1", model_id="model-id", data_subset="all")

    assert job.status == enums.QUEUE_STATUS.INPROGRESS


@responses.activate
def test_training_predictions_job__wait_for_completion__ok():
    resource_url = "https://host_name.com/resource-location/"
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(refresh_job_data),
        content_type="application/json",
        adding_headers={"Location": resource_url},
    )

    job = TrainingPredictionsJob(mock_job_data, model_id="model-id", data_subset="all")
    job.wait_for_completion(max_wait=3)

    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_regression_test_get_completed_pred_with_training_predictions_job(client):

    project_id = "p-id"
    prediction_id = "1"
    data_path = "projects/{project_id}/trainingPredictions/{prediction_id}".format(
        project_id=project_id, prediction_id=prediction_id
    )
    get_predictions_url = "https://host_name.com/{data_path}/".format(data_path=data_path)
    mock_generic_job_data = {
        "status": enums.QUEUE_STATUS.COMPLETED,
        "url": "https://host_name.com/projects/p-id/predictJobs/1/",
        "id": prediction_id,
        "jobType": "a job",
        "isBlocked": False,
        "projectId": project_id,
    }

    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(dict(mock_generic_job_data)),
        content_type="application/json",
        adding_headers={"Location": get_predictions_url},
    )

    job = TrainingPredictionsJob.get(project_id, prediction_id)

    data = job.get_result()
    assert job.status == "COMPLETED"

    assert isinstance(data, TrainingPredictions)
    assert data.project_id == project_id
    assert data.prediction_id == prediction_id
    assert data.path == data_path + "/"



import json

import pytest
import responses

from datarobot import Deployment
from datarobot.utils import from_api


class TestListFeatures(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            "count": 40,
            "next": None,
            "previous": None,
            "data": [
                {
                    "importance": 0.12,
                    "featureType": "Categorical",
                    "knownInAdvance": False,
                    "name": "Feature1",
                },
                {
                    "importance": 0.05,
                    "featureType": "Numeric",
                    "knownInAdvance": True,
                    "name": "Feature2",
                },
            ],
        }

    @pytest.fixture
    def response(self, unittest_endpoint, deployment_data, response_data):
        url = "{}/deployments/{}/features/".format(unittest_endpoint, deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "response")
    def test_retrieve(self, deployment_data, response_data):
        actual_features = Deployment.get(deployment_data["id"]).get_features()
        expected_features = from_api(response_data["data"], keep_null_keys=True)
        assert actual_features == expected_features



import json

import dateutil
import pytest
import responses
import trafaret as t

from datarobot import Deployment
from datarobot.enums import (
    DEPLOYMENT_ACCURACY_HEALTH_STATUS,
    DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE,
    DEPLOYMENT_IMPORTANCE,
    DEPLOYMENT_MODEL_HEALTH_STATUS,
    DEPLOYMENT_SERVICE_HEALTH_STATUS,
)
from datarobot.models.deployment import DeploymentListFilters
from datarobot.utils import from_api
from tests.utils import request_body_to_json


def assert_deployment(deployment, deployment_data):
    def assert_health(health_from_client, health_from_api):
        assert health_from_client["status"] == health_from_client["status"]
        if "startDate" in health_from_api and health_from_api["startDate"]:
            timestamp = health_from_api["startDate"]
            assert health_from_client.get("start_date") == dateutil.parser.parse(timestamp)
        else:
            assert "start_date" not in health_from_client
        if "endDate" in health_from_api and health_from_api["endDate"]:
            timestamp = health_from_api["endDate"]
            assert health_from_client.get("end_date") == dateutil.parser.parse(timestamp)
        else:
            assert "end_date" not in health_from_client

    assert deployment.id == deployment_data["id"]
    assert deployment.label == deployment_data["label"]
    assert deployment.description == deployment_data["description"]

    assert deployment.model == from_api(deployment_data["model"])
    assert deployment.default_prediction_server == from_api(
        deployment_data["defaultPredictionServer"]
    )
    assert deployment.capabilities == from_api(deployment_data["capabilities"])
    assert deployment.permissions == deployment_data["permissions"]

    prediction_usage_from_api = deployment_data["predictionUsage"]
    assert deployment.prediction_usage["daily_rates"] == prediction_usage_from_api["dailyRates"]
    if prediction_usage_from_api["lastTimestamp"]:
        assert deployment.prediction_usage["last_timestamp"] == dateutil.parser.parse(
            prediction_usage_from_api["lastTimestamp"]
        )

    assert_health(deployment.service_health, deployment_data["serviceHealth"])
    assert_health(deployment.model_health, deployment_data["modelHealth"])
    assert_health(deployment.accuracy_health, deployment_data["accuracyHealth"])

    repr = "Deployment({label})".format(label=deployment_data["label"])
    assert str(deployment) == repr


@pytest.fixture()
def deployment_get_response(unittest_endpoint, deployments_data):
    for deployment_data in deployments_data:
        url = "{}/{}/{}/".format(unittest_endpoint, "deployments", deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(deployment_data),
        )


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_get_deployment(deployments_data):
    for deployment_data in deployments_data:
        deployment = Deployment.get(deployment_data["id"])
        assert_deployment(deployment, deployment_data)


class TestDeploymentDelete(object):
    @pytest.fixture()
    def deployment_delete_response(self, unittest_endpoint, deployments_data):
        for deployment_data in deployments_data:
            url = "{}/{}/{}/".format(unittest_endpoint, "deployments", deployment_data["id"])
            responses.add(responses.DELETE, url, status=204)

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "deployment_delete_response")
    def test_delete_deployment(self, deployment_data):
        deployment = Deployment.get(deployment_data["id"])
        deployment.delete()

        assert responses.calls[1].request.method == "DELETE"
        assert responses.calls[1].request.url.endswith("/deployments/5c6f7a9e9ca9b20017ff95a2/")


class TestDeploymentCreateFromLearningModel(object):
    @pytest.fixture
    def learning_model_create_response(self, unittest_endpoint, deployment_data):
        url = "{}/{}/{}/".format(unittest_endpoint, "deployments", "fromLearningModel")
        responses.add(
            responses.POST,
            url,
            body=json.dumps({"id": deployment_data["id"]}),
            status=200,
            content_type="application/json",
        )

    @responses.activate
    @pytest.mark.usefixtures("learning_model_create_response", "deployment_get_response")
    def test_create_minimum(self):
        Deployment.create_from_learning_model("5c76a543962d744efba25b85", "loan is bad")

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body["modelId"] == "5c76a543962d744efba25b85"
        assert create_body["label"] == "loan is bad"
        assert "description" not in create_body
        assert "defaultPredictionServerId" not in create_body

    @responses.activate
    @pytest.mark.usefixtures("learning_model_create_response", "deployment_get_response")
    def test_create_all_options(self):
        Deployment.create_from_learning_model(
            "5c76a543962d744efba25b85",
            "loan is bad",
            "predict if a loan is gonna default",
            "5c0fcdb8962d74370dd0c38e",
        )

        create_body = request_body_to_json(responses.calls[0].request)
        assert create_body["modelId"] == "5c76a543962d744efba25b85"
        assert create_body["label"] == "loan is bad"
        assert create_body["description"] == "predict if a loan is gonna default"
        assert create_body["defaultPredictionServerId"] == "5c0fcdb8962d74370dd0c38e"


@pytest.mark.parametrize("from_version", [True, False])
class TestDeploymentCreateFromCustomModel(object):
    def _make_model_package_get_response(self, get_model_package_url, custom_model_entity_id, data):
        url = "{}?modelId={}".format(get_model_package_url, custom_model_entity_id)
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps({"count": 1, "next": None, "previous": None, "data": data}),
        )

    @pytest.fixture
    def get_model_package_url(self, unittest_endpoint):
        return "{}/modelPackages/".format(unittest_endpoint)

    @pytest.fixture
    def create_model_package_url(self, unittest_endpoint, from_version):
        if from_version:
            route = "fromCustomModelVersion"
        else:
            route = "fromCustomModelImage"
        return "{}/modelPackages/{}/".format(unittest_endpoint, route)

    @pytest.fixture
    def get_deployment_url(self, unittest_endpoint):
        return "{}/deployments/".format(unittest_endpoint)

    @pytest.fixture
    def create_deployment_url(self, unittest_endpoint):
        return "{}/deployments/fromModelPackage/".format(unittest_endpoint)

    @pytest.fixture
    def get_status_url(self, unittest_endpoint):
        return "{}/status_url".format(unittest_endpoint)

    @pytest.fixture
    def custom_model_entity_id(self):
        return "5ea6aeeae8e8a40039714bde"

    @pytest.fixture
    def model_package_data(self):
        return {"id": "5eaaa09a40240308176cee55"}

    @pytest.fixture
    def prediction_server_id(self):
        return "5c0fcdb8962d74370dd0c38e"

    @pytest.fixture
    def model_package_get_response(
        self, get_model_package_url, custom_model_entity_id, model_package_data
    ):
        self._make_model_package_get_response(
            get_model_package_url, custom_model_entity_id, [model_package_data]
        )

    @pytest.fixture
    def model_package_get_response_none(self, get_model_package_url, custom_model_entity_id):
        self._make_model_package_get_response(get_model_package_url, custom_model_entity_id, [])

    @pytest.fixture
    def model_package_create_response(self, create_model_package_url, model_package_data):
        responses.add(
            responses.POST,
            create_model_package_url,
            body=json.dumps(model_package_data),
            status=200,
            content_type="application/json",
        )

    @pytest.fixture
    def deployment_create_response(
        self, create_deployment_url, get_deployment_url, get_status_url, deployment_data
    ):
        responses.add(
            responses.POST,
            create_deployment_url,
            headers={"Location": get_status_url},
            body=json.dumps({"id": deployment_data["id"]}),
            status=200,
            content_type="application/json",
        )
        responses.add(
            responses.GET, get_status_url, headers={"Location": get_deployment_url}, status=303
        )

    @pytest.fixture
    def _assert_deployment_creation(
        self,
        model_package_data,
        deployment_data,
        prediction_server_id,
        create_deployment_url,
        get_deployment_url,
        get_status_url,
    ):
        def _assert(
            create_deployment_req, get_status_req, get_deployment_req,
        ):
            assert create_deployment_req.method == "POST"
            assert create_deployment_req.url == create_deployment_url
            create_body = request_body_to_json(create_deployment_req)
            assert set(create_body) == {
                "modelPackageId",
                "label",
                "description",
                "defaultPredictionServerId",
            }
            assert create_body["modelPackageId"] == model_package_data["id"]
            assert create_body["label"] == "My deployment"
            assert create_body["description"] == "deployment description"
            assert create_body["defaultPredictionServerId"] == prediction_server_id

            assert get_status_req.method == "GET"
            assert get_status_req.url == get_status_url

            assert get_deployment_req.method == "GET"
            assert get_deployment_req.url == "{}{}/".format(
                get_deployment_url, deployment_data["id"]
            )

        return _assert

    @responses.activate
    @pytest.mark.usefixtures(
        "model_package_get_response", "deployment_create_response", "deployment_get_response"
    )
    def test_create_deployment_model_package_exists(
        self,
        custom_model_entity_id,
        prediction_server_id,
        get_model_package_url,
        _assert_deployment_creation,
        from_version,
    ):
        args = [
            custom_model_entity_id,
            "My deployment",
            "deployment description",
            prediction_server_id,
        ]
        if from_version:
            Deployment.create_from_custom_model_version(*args)
        else:
            Deployment.create_from_custom_model_image(*args)

        assert len(responses.calls) == 4

        responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == "{}?modelId={}".format(
            get_model_package_url, custom_model_entity_id
        )

        create_deployment_req = responses.calls[1].request
        get_status_req = responses.calls[2].request
        get_deployment_req = responses.calls[3].request

        _assert_deployment_creation(create_deployment_req, get_status_req, get_deployment_req)

    @responses.activate
    @pytest.mark.usefixtures(
        "model_package_get_response_none",
        "model_package_create_response",
        "deployment_create_response",
        "deployment_get_response",
    )
    def test_create_deployment_model_package_doesnt_exist(
        self,
        custom_model_entity_id,
        prediction_server_id,
        get_model_package_url,
        create_model_package_url,
        _assert_deployment_creation,
        from_version,
    ):
        args = [
            custom_model_entity_id,
            "My deployment",
            "deployment description",
            prediction_server_id,
        ]
        if from_version:
            Deployment.create_from_custom_model_version(*args)
        else:
            Deployment.create_from_custom_model_image(*args)

        assert len(responses.calls) == 5

        # get model package
        assert responses.calls[0].request.method == "GET"
        assert responses.calls[0].request.url == "{}?modelId={}".format(
            get_model_package_url, custom_model_entity_id
        )

        # create model package
        assert responses.calls[1].request.method == "POST"
        assert responses.calls[1].request.url == create_model_package_url
        create_body = request_body_to_json(responses.calls[1].request)
        expected_field = "customModelVersionId" if from_version else "customModelImageId"
        assert set(create_body) == {expected_field}
        assert create_body[expected_field] == custom_model_entity_id

        create_deployment_req = responses.calls[2].request
        get_status_req = responses.calls[3].request
        get_deployment_req = responses.calls[4].request

        _assert_deployment_creation(create_deployment_req, get_status_req, get_deployment_req)


class TestDeploymentList(object):
    @pytest.fixture()
    def deployments_list_response_data(self, deployments_data):
        return {"count": 1, "next": None, "previous": None, "data": deployments_data}

    @pytest.fixture()
    def deployment_list_response(self, unittest_endpoint, deployments_list_response_data):
        url = "{}/{}/".format(unittest_endpoint, "deployments")
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(deployments_list_response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_list_response")
    def test_list_deployment(self, deployments_data):
        deployments_data = {item["id"]: item for item in deployments_data}
        deployments = Deployment.list()
        assert len(deployments) == len(deployments_data)
        for deployment in deployments:
            assert_deployment(deployment, deployments_data[deployment.id])

    @responses.activate
    @pytest.mark.usefixtures("deployment_list_response")
    def test_sort(self):
        Deployment.list(order_by="-label")
        assert "deployments/?orderBy=-label" in responses.calls[0].request.url

    @responses.activate
    @pytest.mark.usefixtures("deployment_list_response")
    def test_search(self):
        Deployment.list(search="readmitted")
        assert "deployments/?search=readmitted" in responses.calls[0].request.url

    @responses.activate
    @pytest.mark.usefixtures("deployment_list_response")
    @pytest.mark.parametrize(
        "filter_name,filter_value,query_param_key,query_param_value",
        [
            ("role", "OWNER", "role", "OWNER"),
            (
                "service_health",
                DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING,
                "serviceHealth",
                "passing",
            ),
            (
                "service_health",
                [DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING],
                "serviceHealth",
                "passing",
            ),
            (
                "service_health",
                [
                    DEPLOYMENT_SERVICE_HEALTH_STATUS.FAILING,
                    DEPLOYMENT_SERVICE_HEALTH_STATUS.WARNING,
                ],
                "serviceHealth",
                "failing,warning",
            ),
            ("model_health", DEPLOYMENT_MODEL_HEALTH_STATUS.PASSING, "modelHealth", "passing"),
            ("model_health", [DEPLOYMENT_MODEL_HEALTH_STATUS.FAILING], "modelHealth", "failing"),
            (
                "model_health",
                [DEPLOYMENT_MODEL_HEALTH_STATUS.PASSING, DEPLOYMENT_MODEL_HEALTH_STATUS.UNKNOWN],
                "modelHealth",
                "passing,unknown",
            ),
            (
                "accuracy_health",
                DEPLOYMENT_ACCURACY_HEALTH_STATUS.UNAVAILABLE,
                "accuracyHealth",
                "unavailable",
            ),
            (
                "accuracy_health",
                [DEPLOYMENT_ACCURACY_HEALTH_STATUS.WARNING],
                "accuracyHealth",
                "warning",
            ),
            (
                "accuracy_health",
                [
                    DEPLOYMENT_ACCURACY_HEALTH_STATUS.FAILING,
                    DEPLOYMENT_ACCURACY_HEALTH_STATUS.PASSING,
                ],
                "accuracyHealth",
                "failing,passing",
            ),
            (
                "importance",
                [DEPLOYMENT_IMPORTANCE.HIGH, DEPLOYMENT_IMPORTANCE.LOW],
                "importance",
                "HIGH,LOW",
            ),
        ],
    )
    def test_filters(self, filter_name, filter_value, query_param_key, query_param_value):
        filter_args = {filter_name: filter_value}
        filters = DeploymentListFilters(**filter_args)
        Deployment.list(filters=filters)

        expected_call = "deployments/?{}={}".format(
            query_param_key, query_param_value.replace(",", "%2C")
        )

        actual_call = responses.calls[0].request.url
        assert actual_call.endswith(
            expected_call
        ), "Expected called url to end with: {}\n Got: {}".format(expected_call, actual_call)

    @responses.activate
    @pytest.mark.usefixtures("deployment_list_response")
    def test_all_query_params(self):
        Deployment.list(
            order_by="-label",
            search="readmitted",
            filters=DeploymentListFilters(
                role="OWNER",
                service_health=DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING,
                model_health=DEPLOYMENT_MODEL_HEALTH_STATUS.WARNING,
                accuracy_health=DEPLOYMENT_ACCURACY_HEALTH_STATUS.FAILING,
                execution_environment_type=DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE.DATAROBOT,
            ),
        )
        expected_args = [
            "orderBy=-label",
            "search=readmitted",
            "role=OWNER",
            "serviceHealth=passing",
            "modelHealth=warning",
            "accuracyHealth=failing",
            "executionEnvironmentType=datarobot",
        ]
        query_param_str = responses.calls[0].request.url.split("deployments/?")[1]
        query_params = query_param_str.split("&")

        assert len(expected_args) == len(query_params)

        for expected_arg in expected_args:
            assert (
                expected_arg in query_params
            ), "Expected to find the query arg {}, but didn't.\n Got: {}".format(
                expected_arg, query_param_str
            )


class TestDeploymentUpdate(object):
    @pytest.fixture()
    def deployment_update_response(self, unittest_endpoint, deployments_data):
        for deployment_data in deployments_data:
            url = "{}/{}/{}/".format(unittest_endpoint, "deployments", deployment_data["id"])
            responses.add(responses.PATCH, url, status=204)

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "deployment_update_response")
    def test_cannot_update_nothing(self, deployment_data):
        deployment = Deployment.get(deployment_data["id"])
        with pytest.raises(ValueError):
            deployment.update()

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "deployment_update_response")
    def test_update_label(self, deployment_data):
        deployment = Deployment.get(deployment_data["id"])
        deployment.update(label="new label")

        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url.endswith("/deployments/5c6f7a9e9ca9b20017ff95a2/")

        request_json = json.loads(responses.calls[1].request.body)
        assert request_json["label"] == "new label"

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "deployment_update_response")
    def test_update_description(self, deployment_data):
        deployment = Deployment.get(deployment_data["id"])
        deployment.update(description="new description")

        assert responses.calls[1].request.method == "PATCH"
        assert responses.calls[1].request.url.endswith("/deployments/5c6f7a9e9ca9b20017ff95a2/")

        request_json = json.loads(responses.calls[1].request.body)
        assert request_json["description"] == "new description"


class TestDeploymentListFilters(object):
    @pytest.mark.parametrize(
        "filter_name,valid_value",
        [
            ("role", "OWNER"),
            ("role", "USER"),
            ("service_health", DEPLOYMENT_SERVICE_HEALTH_STATUS.PASSING),
            ("service_health", [DEPLOYMENT_SERVICE_HEALTH_STATUS.UNKNOWN]),
            ("service_health", DEPLOYMENT_SERVICE_HEALTH_STATUS.ALL),
            ("model_health", DEPLOYMENT_MODEL_HEALTH_STATUS.FAILING),
            ("model_health", [DEPLOYMENT_MODEL_HEALTH_STATUS.WARNING]),
            ("model_health", DEPLOYMENT_MODEL_HEALTH_STATUS.ALL),
            ("accuracy_health", DEPLOYMENT_ACCURACY_HEALTH_STATUS.UNAVAILABLE),
            ("accuracy_health", [DEPLOYMENT_ACCURACY_HEALTH_STATUS.UNKNOWN]),
            ("accuracy_health", DEPLOYMENT_ACCURACY_HEALTH_STATUS.ALL),
            ("execution_environment_type", DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE.DATAROBOT),
            ("execution_environment_type", DEPLOYMENT_EXECUTION_ENVIRONMENT_TYPE.ALL),
            ("importance", DEPLOYMENT_IMPORTANCE.MODERATE),
            ("importance", DEPLOYMENT_IMPORTANCE.ALL),
        ],
    )
    def test_valid_filter_values(self, filter_name, valid_value):
        kwargs = {filter_name: valid_value}
        DeploymentListFilters(**kwargs)

    @pytest.mark.parametrize(
        "filter_name,invalid_value",
        [
            ("role", 1234),
            ("service_health", 1234),
            ("model_health", 1234),
            ("accuracy_health", 1234),
            ("execution_environment_type", 1234),
            ("importance", 1234),
        ],
    )
    def test_invalid_filter_values(self, filter_name, invalid_value):
        with pytest.raises(t.DataError):
            kwargs = {filter_name: invalid_value}
            DeploymentListFilters(**kwargs)



import json

import pytest
import responses

from datarobot import errors


@pytest.fixture
def cv_scores_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/crossValidationScores/".format(
        project_id, model_id
    )


@pytest.fixture
def cv_scores():
    return {
        "metric1": {"1.0": "val", "2.0": "val2"},
        "metric2": {"1.0": "val3", "2.0": "val4"},
    }


@pytest.fixture
def cv_scores_get_response(cv_scores_url, cv_scores):
    responses.add(
        responses.GET,
        cv_scores_url,
        status=200,
        content_type="applicaiton/json",
        body=json.dumps(cv_scores),
    )


@pytest.fixture
def cv_scores_get_response_404(cv_scores_url, cv_scores):
    responses.add(
        responses.GET,
        cv_scores_url,
        status=404,
        content_type="applicaiton/json",
        body="object does not contain cross validation scores",
    )


@pytest.fixture
def cv_scores_get_response_filter_by_metric(cv_scores_url, cv_scores):
    url = cv_scores_url + "?metric=metric1"
    responses.add(
        responses.GET, url, status=200, content_type="applicaiton/json", body=json.dumps(cv_scores)
    )


@pytest.fixture
def cv_scores_get_response_filter_by_partition(cv_scores_url, cv_scores):
    url = cv_scores_url + "?partition=2"
    responses.add(
        responses.GET, url, status=200, content_type="applicaiton/json", body=json.dumps(cv_scores)
    )


@pytest.fixture
def cv_scores_get_response_filter_by_partition_and_metric(cv_scores_url, cv_scores):
    url = cv_scores_url + "?partition=2&metric=metric1"
    responses.add(
        responses.GET, url, status=200, content_type="applicaiton/json", body=json.dumps(cv_scores)
    )


@responses.activate
@pytest.mark.usefixtures("cv_scores_get_response")
def test_get_cross_validation_scores(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores()
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures("cv_scores_get_response_filter_by_metric")
def test_get_cross_validation_scores_filtered_by_metric(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores(metric="metric1")
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures("cv_scores_get_response_filter_by_partition")
def test_get_cross_validation_scores_filtered_by_partition(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores(partition=2)
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures("cv_scores_get_response_filter_by_partition_and_metric")
def test_get_cross_validation_scores_filtered_by_partition_and_metric(one_model, cv_scores):
    cv_scores_retrieved = one_model.get_cross_validation_scores(partition=2, metric="metric1")
    assert cv_scores_retrieved == cv_scores


@responses.activate
@pytest.mark.usefixtures("cv_scores_get_response_404")
def test_get_cross_validation_scores_404(one_model):
    with pytest.raises(errors.ClientError):
        one_model.get_cross_validation_scores()
